/********************************************************************/
/* wiczenie 8 - Zdalna regulacja obrotw silnika DC z komputera PC */
/*               Modulacja PWM przy uyciu timera1                  */
/*               Wykorzystanie UART-a do transmisji z komputerem    */
/* J.D. '2003                                                       */
/********************************************************************/

#include <io.h>
#include <progmem.h>
#include <stdlib.h>
#include <interrupt.h>
#include <signal.h>

#define FCPU      8000000    //czstotliwo oscylatora CPU
#define VUART     38400      //prdko transmisji [bit/s]
#define VUBRR     FCPU/(VUART*16)-1    //wpis do UBRR dla VUART

unsigned char romram;        //romram=1 => dane z pamici programu
                             //romram=0 => dane z RAM-u
char *pfifosio;              //wskanik na kolejk UART-u
unsigned char volatile fodbznak=0;    //flaga: "odebrano znak"
char komenda;                //odebrana komenda z PC-ta
char *fifosio[];            //wskanik na kolejk UART-u


SIGNAL(SIG_UART_RECV)        //procedura obsugi odbiornika UART-u
{
 komenda=UDR;                //zapamitaj odebran komend
 fodbznak=1;                 //ustaw flag odebrania znaku
}

SIGNAL(SIG_UART_TRANS)       //procedura obsugi nadajnika UART
{                            //wywoywana po wysaniu znaku
 char znak;

 if(romram)                  //skd pobiera dane?
 {
  znak=PRG_RDB(pfifosio++); //pobierz dan z pamici programu
 }
 else
 {
  znak=*pfifosio++;         //pobierz dane z pamici RAM
 }
 if(znak!=0)                 //czy koniec pobierania danych?
 {
  UDR=znak;                  //nie, wylij znak pobrany z kolejki
 }
 else
 {
  cbi(UCR,TXEN);             //tak, wycz nadajnik
 }
}

void czekaj(unsigned long zt) //procedura wytracania czasu
{
 #define tau 10.38
 unsigned char zt1;
 for(;zt>0;zt--)
 {
  for(zt1=255;zt1!=0;zt1--);
 }
}

void wyslijtekstROM(char *tekst)  //wysyanie danych z pamici programu
{
 romram=1;                   //dane bd z pamici programu
 pfifosio=tekst;            //ustaw wskanik na dane do wysania
 sbi(UCR,TXEN);              //wcz nadajnik
 UDR=PRG_RDB(pfifosio++);   //wylij pierwszy znak, pozostae bd pobierane
                             //w procedurze obsugi przerwania TXC
}

void wyslijtekst(char *tekst)     //wysyanie danych z pamici programu
{
 romram=0;                   //dane bd z pamici danych
 pfifosio=tekst;            //ustaw wskanik na dane do wysania
 sbi(UCR,TXEN);              //wcz nadajnik
 UDR=*pfifosio++;           //wylij pierwszy znak, pozostae bd pobierane
                             //w procedurze obsugi przerwania TXC
}

int main(void)               
{
 unsigned char i;
 unsigned char volatile licznikkl=0;  //zmienna wykorzystywana do pomiaru czasu
                             //nacinicia przyciskw

 char volatile przyrost=1;  //przyrost zmiany wspczynnika wypenienia sygnau PWM
 //tablica komunikatw do wysania
 char *info[7]={
                PSTR("\n\rRegulator obrotw silnika DC\n\r"),
                PSTR(", - zmniejszanie obrotw\n\r"),
                PSTR(". - zwikszanie obrotw\n\r"),
                PSTR("0 - zatrzymanie silnika\n\r"),
                PSTR("1 - start z max. obrotami\n\r"),
                PSTR("N - podaj aktualne parametry sterownika\n\r\n"),
                PSTR("\n\rAktualne parametry PWM:")
               };

 union                       //unia pozwala na bajtowy dostp do zmiennej int
      {
       unsigned int pwm;
       unsigned char pwmc[2];
      }volatile upwm;        //aktualny wspczynnik wypenienia sygn. PWM

 DDRB=0xff;    //PORTB - wy
 PORTB=0xff;   
 DDRD=0x02;    //PD1 - wy (RXD), pozostae we
 PORTD=0x02;   //podcigania wejcia PD1 (RXD)
 UBRR=VUBRR;   //ustaw prdko transmisji
 UCR=1<<RXCIE | 1<<TXCIE  | 1<<RXEN;   //zezwolenie na przerwania od
                   //odbiornika i nadajnika, zezwolenie na odbir i nadawanie
 TCCR1A=0x83;      //PWM 10 bitowy
                   //zerowanie OC1 po spenieniu warunku rwnoci podczas liczenia
                   //w gr, ustawiane podczas liczenia w d
 TCCR1B=0x01;      //preskaler=3, co przy 10-bit PWM daje Fwy=ok. 61Hz @8MHz
 TCNT1L=0x00;      //wstpne ustawienie licznika 1
 TCNT1H=0x00;
 upwm.pwm=0x3ff;   //pocztkowo silnik wczony, warto TOP odpowiada wysokiemu
                   //poziomowi na wyjciu OC1 (PB3)
 OCR1H=upwm.pwmc[1];         //wpisz aktualnie ustawiony wspczynnik do rejestrw
 OCR1L=upwm.pwmc[0];         //OCR1 timera1
 sei();            //wcz przerwania

 for(i=0;i<5;i++) //wylij winietk
 {
  wyslijtekstROM(info[i]);   //wysanie pojedynczej linii tekstu
  while(bit_is_set(UCR,TXEN));    //trzeba zaczeka, a zostanie wysana do koca
 }

 while(1)          //gwna ptla programu
 {
  if(fodbznak)          //czy odebrano jaki znak?
  {
   fodbznak=0;          //tak
   switch (komenda)     //interpretacja komendy i wykonanie odpowiedniej akcji
   {
    case '.':                 //odebrano "." - zwiksz prdko
              upwm.pwm+=przyrost;       //zwiksz PWM
              if(upwm.pwm>0x3ff)
              {
               upwm.pwm=0x3ff;    //jeli przekroczono warto TOP, to ustaw TOP
              }
              czekaj(150*tau);    //eliminacja powtrnej interpretacji
                                  //nacinicia przycisku
              licznikkl++;              //mierz dugo nacinicia przycisku
              break;
    case ',':                 //odebrano "," - zmniejsz prdko
              upwm.pwm-=przyrost;      //zmniejsz PWM
              if(upwm.pwm>0x3ff)
              {
               upwm.pwm=0;        //jeli przekroczono warto zero to ustaw zero
              }
              czekaj(150*tau);    //eliminacja powtrnej interpretacji
                                  //nacinicia przycisku 
              licznikkl++;        //mierz dugo nacinicia przycisku
              break;
    case '0':                //odebrano "0" - zatrzymaj silnik
              upwm.pwm=0;    //silnik STOP
              break;
    case '1':                //odebrano "1" - ustaw max obroty silnika
              upwm.pwm=0x3ff;     //silnik na MAX
              break;
    case 'n':
    case 'N':
              wyslijtekstROM(info[6]);      //wysanie pojedynczej linii tekstu
              while(bit_is_set(UCR,TXEN));  //trzeba zaczeka, 
                                            //a zostanie wysana do koca
              wyslijtekst("0x");       //wylij prefiks dla liczb heksadecymalnych
              while(bit_is_set(UCR,TXEN));  //trzeba zaczeka, 
                                            //a zostanie wysana do koca
              itoa(upwm.pwm,fifosio,16);   //konwersja liczby int (hex)
                                            //na acuch znakowy
              wyslijtekst(fifosio);        //wylij aktualn warto PWM do PC-ta
              while(bit_is_set(UCR,TXEN));  //trzeba zaczeka,
                                            //a zostanie wysana do koca
              break;
   }
   if(licznikkl>6)
   {
    przyrost=+16;             //wykryto dugie nacinicie klawisza,
                              //zwiksz krok regulacji
    licznikkl=6;              //dalej ju nie zwikszaj kroku
    cbi(PORTB,1);             //zapal diod LED2
   }
   OCR1H=upwm.pwmc[1];        //wpisz aktualnie ustawiony wspczynnik do rejestrw
   OCR1L=upwm.pwmc[0];        //OCR1 timera1
  }
  else
  {                           //jeli cisza na linii, ustaw parametry spoczynkowe
   licznikkl=0;
   przyrost=1;
   sbi(PORTB,1);              //zga diod LED2
  }
 }
}

