/***************************************************************/
/* wiczenie 4a - sterowanie alfanumerycznym wywietlaczem LCD */
/*                16x2 (16 znakw, 2 wiersze)                  */
/*              - obsuga pojedynczego klawisza                */
/* J.D. '2003                                                  */
/***************************************************************/
#include <io.h>
#include <progmem.h>
#include <stdlib.h>

#define lcd_rs 2             //definicja bitu portu dla linii RS
#define lcd_e 3              //definicja bitu portu dla linii E
#define CR 0x0a              //definicja znaku CR (przejcie do nowej linii)

unsigned char wiersz=0;
unsigned char kolumna=0;

void czekaj(unsigned long pt) //procedura wytracania czasu
{                   
#define tau 10.38             //przybliony przelicznik argumentu na ms
 unsigned char tp1;

 for(;pt>0;pt--)
 {
  for(tp1=255;tp1!=0;tp1--);
 }
}

void piszilcd(unsigned char instr) //zapisz instrukcj sterujc do LCD
{
 cbi(PORTB,lcd_rs);
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|(instr&0xf0);  //przygotuj starszy pbajt do LCD
 asm("nop");                       //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD ok. 100us
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|((instr&0x0f)<<4);  //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD ok. 100us
}

void piszdlcd(char dana)           //zapisz dan do LCD
{
 sbi(PORTB,lcd_rs);
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|(dana&0xf0);   //przygotuj starszy pbajt do LCD
 asm("nop");                       //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|((dana&0x0f)<<4); //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD
}

void czysclcd(void)                //czy ekran
{
 piszilcd(0x01);      //polecenie czyszczenia ekranu dla kontrolera LCD
 czekaj(1.64*tau);    //rozkaz 0x01 wykonuje si 1.64ms
 wiersz=0;
 kolumna=0;
}

void piszznak(char znak)    //procedura umieszcza znak na wywietlaczu
{
 piszdlcd(znak);            //wywietl znak na LCD
 if(++kolumna==16)          //czy bieca kolumna mieci si na wywietlaczu?
 {                          
  kolumna=0;                //jeli nie, to ustaw pocztkow...
  if(++wiersz==2)           //i przejd do nowego wiersza
  {
   wiersz=0;       //jeli nowy wiersz jest poza wywietlaczem, ustaw pocztkowy
  }
 }
}

void lcdxy(unsigned char w, unsigned char k)    //ustaw wsprzdne kursora
{
 piszilcd((w*0x40+k)|0x80);  //standardowy rozkaz sterownika LCD 
                             //ustawiajcy kursor w okrelonych wsprzdnych
}

void pisztekst(char *tekst)  //pisz tekst na LCD wskazywany pointerem *tekst
{
 char zn;
 char nr=0;

 while(1)
 {
  zn=PRG_RDB(&tekst[nr++]);  //pobranie znaku z pamici programu
  if(zn!=0)                  //czy nie ma koca tekstu?
  {
   if(zn==CR)                //czy znak nowej linii
   {
    wiersz==1?wiersz=0:++wiersz; //przejd do nowej linii
    kolumna=0;
    lcdxy(wiersz,kolumna);   //ustaw obowizujce po zmianie wsprzdne na LCD
   }
   else
   {
    piszdlcd(zn);            //umie pojedynczy znak tekstu na LCD
   }
  }
  else
  {
   break;                    //zakocz ptl, jeli koniec tekstu
  }
 }
}

void klawisz(unsigned char pozkl)  //czekaj na nacinicie okrelonego klawisza
                                   //klawisz jest wybierany poprzez argument pozkl
{
 while(PIND&pozkl);                //czekaj na nacinicie klawisza
 czekaj(30*tau);                   //odczekaj a ustan zakcenia
 while(!(PIND&pozkl));             //czekaj na zwolnienie klawisza
}

int main(void) //program gwny
{
 unsigned char i;
 //>>>>>>>>>>>>>    definicje tekstw wywietlanych na LCD   <<<<<<<<<<<<<<
 char *tekst1=PSTR("kursor, mruga");
 char *tekst2=PSTR("bez kursora\nbez mrugania");  
                                   // \n oznacza znak CR (przejcie do nowej linii)
 char *tekst3=PSTR("kursor\nbez mrugania");
 char *tekst4=PSTR("bez kursora\nmruganie");
 char *tekst5=PSTR("Przesuwanie\ntekstu w prawo");
 char *tekst6=PSTR("Przesuwanie\ntekstu w lewo");
 char *tekst7=PSTR("pozycja 1,3");
 char *tekst8=PSTR("pozycja 0,9");
 char *tekst9=PSTR("Hello World!");

 //>>>>>>>>>>>>>>>>>>>>    konfigurowanie portw <<<<<<<<<<<<<<<<<<<<<<<<<<
 PORTD=0xff;   //port z podciganiem
 PORTB=0x03;   //port z podciganiem
 DDRD=0x00; //PORTD - we
 DDRB=0xff; //PORTB - wy

 //>>>>>>>>>>>>>>>>>>   inicjacja wywietlacza LCD   <<<<<<<<<<<<<<<<<<<<<<
 czekaj(45*tau);        //opnienie ok. 45ms dla ustabilizowania si napicia
                        //zasilania LCD (katalogowo min. 15 ms)
                        //lcd_rs ju wczeniej byo ustawione w stanie "0"
 for(i=0;i<3;i++)       //3-krotne wysanie 3-
 {
  sbi(PORTB,lcd_e);
  PORTB=(PORTB&0x0f)|0x30; //wylij 3- do LCD
  asm("nop");
  asm("nop");
  asm("nop");
  cbi(PORTB,lcd_e);
  czekaj(5*tau);      //ok. 5ms
 }
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|0x20;  //wylij 2- do LCD
 asm("nop");               //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);
 czekaj(10L);        //od tego momentu mona sprawdza gotowo LCD
                     //w tym programie nie bdzie sprawdzania gotowoci

 piszilcd(0x28);     //interfejs 4-bitowy, 2 linie, znak 5x7
 piszilcd(0x08);     //wycz LCD, wycz kursor, wycz mruganie
 piszilcd(0x01);     //czy LCD
 czekaj(1.64*tau);   //wymagane dla instrukcji czyszczenia ekranu opnienie
 piszilcd(0x06);     //bez przesuwania w prawo

 while(1)
 {
//>>>>>>>>>>>>   EFEKT 1   <<<<<<<<<<<<<<<
  piszilcd(0x0f);    //wcz LCD, wcz kursor, wcz mruganie
  pisztekst(tekst1);
  klawisz(2);
  czysclcd();

//>>>>>>>>>>>>   EFEKT 2   <<<<<<<<<<<<<<<
  piszilcd(0x0c);    //wcz LCD, bez kursora, bez mrugania
  pisztekst(tekst2);
  klawisz(2);
  czysclcd();

//>>>>>>>>>>>>   EFEKT 3   <<<<<<<<<<<<<<<
  piszilcd(0x0e);    //wcz LCD, wcz kursor, bez mrugania
  pisztekst(tekst3);
  klawisz(2);
  czysclcd();

//>>>>>>>>>>>>   EFEKT 4   <<<<<<<<<<<<<<<
  piszilcd(0x0d);    //wcz LCD, bez kursora, z mruganiem
  pisztekst(tekst4);
  klawisz(2);
  czysclcd();

//>>>>>>>>>>>>   EFEKT 5   <<<<<<<<<<<<<<<
  pisztekst(tekst5); //umie tekst na LCD, a nastpnie przesu w prawo
  czekaj(1000L*tau);
  for(i=0;i<16;i++)  
  {
   piszilcd(0x1c);   //przesu
   czekaj(200*tau);
  }
  czysclcd();

//>>>>>>>>>>>>   EFEKT 6   <<<<<<<<<<<<<<<
  pisztekst(tekst6);  //umie tekst na LCD, a nastpnie przesu w lewo
  czekaj(1000L*tau);
  for(i=0;i<16;i++)  
  {
   piszilcd(0x18);   //przesu
   czekaj(200*tau);
  }
  czysclcd();

//>>>>>>>>>>>>   EFEKT 7   <<<<<<<<<<<<<<<
//umieszczanie kursora w rnych wsprzdnych ekranu
  pisztekst(tekst7);  
  lcdxy(1,3);
  klawisz(2);
  czysclcd();

  pisztekst(tekst8);
  lcdxy(0,9);
  klawisz(2);
  czekaj(500*tau);

//>>>>>>>>>>>>   EFEKT 8   <<<<<<<<<<<<<<<
//wprowadzanie tekstu spoza ekranu
  piszilcd(0x0c);      //wycz kursor
  czysclcd();
  piszilcd(0x80|0x10); //ustaw wsprzdne poza ekranem
  pisztekst(tekst9);   //pisz tekst poza ekranem...
  for(i=0;i<14;i++)    //i przesu w lewo
  {
   piszilcd(0x18);     //przesu
   czekaj(200*tau);
  }
  czekaj(200*tau);

//>>>>>>>>>>>>   EFEKT 9   <<<<<<<<<<<<<<<
//mruganie tekstem wywietlanym na LCD
//ze zmienn czstotliwoci
  for(i=10;i!=0xff;i--)
  {
   piszilcd(0x08);   //wycz LCD
   czekaj(i*1038);
   piszilcd(0x0c);   //wcz LCD
   czekaj(i*1038);
  }
  czekaj(200*tau);
  klawisz(2);

  czysclcd();
  lcdxy(1,2);        //obsunicie tekstu w d
  pisztekst(tekst9);
  czekaj(500*tau);
  for(i=0;i<16;i++)  //zatarcie tekstu znak po znaku
  {
   lcdxy(1,i);
   piszznak(' ');
   czekaj(200*tau);
  }
  czysclcd();
 }
}
