/***************************************************************/
/* wiczenie 4b - sterowanie alfanumerycznym wywietlaczem LCD */
/*                16x1 (16 znakw, 1 wiersz)                   */
/*              - obsuga pojedynczego klawisza                */
/* J.D. '2003                                                  */
/***************************************************************/
#include <io.h>
#include <progmem.h>

#define lcd_rs 2
#define lcd_e 3
#define CR 0x0a

volatile unsigned char kolumna;

void czekaj(unsigned long pt) //procedura wytracania czasu
{
#define tau 10.38             //przybliony przelicznik argumentu na ms
 unsigned char tp1;

 for(;pt>0;pt--)
 {
  for(tp1=255;tp1!=0;tp1--);
 }
}

void piszilcd(unsigned char instr) //zapisz instrukcj sterujc do LCD
{
 cbi(PORTB,lcd_rs);
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|(instr&0xf0);  //przygotuj starszy pbajt do LCD
 asm("nop");                       //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD ok. 100us
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|((instr&0x0f)<<4);  //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD ok. 100us
}

void piszdlcd(char dana)           //zapisz dan do LCD
{
 sbi(PORTB,lcd_rs);
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|(dana&0xf0);   //przygotuj starszy pbajt do LCD
 asm("nop");                       //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|((dana&0x0f)<<4); //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD
}

void czysclcd(void)                //czy ekran
{
 piszilcd(0x01);      //polecenie czyszczenia ekranu dla kontrolera LCD
 czekaj(1.64*tau);    //rozkaz 0x01 wykonuje si 1.64ms
 kolumna=0;
}

void piszznak(char znak)    //procedura umieszcza znak na wywietlaczu
{
 unsigned char p;
 if(kolumna<8)
 {
  p=kolumna;
 }
 else
 {
  p=kolumna+56;       //dodatkowe przesunicie wsprzdnych dla LCD 1x16
                      //dla drugiej powki wywietlacza
 }
 piszilcd(p|0x80);
 kolumna=kolumna!=15?++kolumna:0;  //ustaw nastpn kolumn
 piszdlcd(znak);                   //wywietl znak na LCD
}

void lcdy(unsigned char k)  //ustaw wsprzdne kursora
{
 if(k<8)
 {
  piszilcd(k|0x80);         //standardowy rozkaz sterownika LCD
                            //ustawiajcy kursor w okrelonych wsprzdnych
 }
 else
 {
   piszilcd((k+56)|0x80);   //dodatkowe przesunicie wsprzdnych dla LCD 1x16
 }
 kolumna=k;
}

void pisztekst(char *tekst)  //pisz tekst na LCD wskazywany pointerem *tekst
{
 char nr=0;
 char zn;

 while(1)
 {
  zn=PRG_RDB(&tekst[nr++]);  //pobranie znaku z pamici programu
  if(zn!=0)                  //czy nie ma koca tekstu?
  {
   if(zn==CR)                //czy znak nowej linii
   {
    kolumna=0;       //jeli tak, ustaw obowizujce po zmianie wsprzdne na LCD
   }
   else
   {
    piszznak(zn);            //umie pojedynczy znak tekstu na LCD
   }
  }
  else
  {
   break;                    //zakocz ptl, jeli koniec tekstu
  }
 }
}

void klawisz(unsigned char pozkl)  //czekaj na nacinicie okrelonego klawisza
                                   //klawisz jest wybierany poprzez argument pozkl
{
 while((PIND&pozkl)==pozkl);       //czekaj na nacinicie klawisza
 czekaj(30*tau);                   //odczekaj a ustan zakcenia
 while((PIND&pozkl)!=pozkl);       //czekaj na zwolnienie klawisza
}

int main(void) //program gwny
{
 unsigned char i;

 //>>>>>>>>>>>>>    definicje tekstw wywietlanych na LCD   <<<<<<<<<<<<<<
 char *tekst1=PSTR("Wyswietlacz 1x16");
 char *tekst2=PSTR("Hello World!");

 //>>>>>>>>>>>>>>>>>>>>    konfigurowanie portw <<<<<<<<<<<<<<<<<<<<<<<<<<
 PORTD=0xff;   //port z podciganiem
 PORTB=0x03;   //port z podciganiem
 DDRD=0x00;    //PORTD - we
 DDRB=0xff;    //PORTB - wy

 //>>>>>>>>>>>>>>>>>>   inicjacja wywietlacza LCD   <<<<<<<<<<<<<<<<<<<<<<
 czekaj(45*tau);        //opnienie ok. 45ms dla ustabilizowania si napicia
                        //zasilania LCD (katalogowo min. 15 ms)
                        //lcd_rs ju wczeniej byo ustawione w stanie "0"
 for(i=0;i<3;i++)       //3-krotne wysanie 3-
 {
  sbi(PORTB,lcd_e);
  PORTB=(PORTB&0x0f)|0x30; //wylij 3- do LCD
  asm("nop");
  asm("nop");
  asm("nop");
  cbi(PORTB,lcd_e);        //impuls strobujcy
  czekaj(5*tau);           //ok. 5ms
 }
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|0x20;  //wylij 2- do LCD
 asm("nop");               //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);         //impuls strobujcy
 czekaj(10L);        
 piszilcd(0x28);     //interfejs 4-bitowy, 2 linie, znak 5x7
 piszilcd(0x0c);     //wycz LCD, wycz kursor, wycz mruganie
 piszilcd(0x01);     //czy LCD
 czekaj(1.64*tau);   //wymagane dla instrukcji czyszczenia ekranu opnienie
 piszilcd(0x06);     //bez przesuwania w prawo

 while(1)
 {
//>>>>>>>>>>>>   EFEKT 1   <<<<<<<<<<<<<<<
//wywietl tekst
  pisztekst(tekst1);
  klawisz(2);
  pisztekst(tekst2);
  klawisz(2);
  czysclcd();

//>>>>>>>>>>>>   EFEKT 2   <<<<<<<<<<<<<<<
//umieszczanie kursora w rnych wsprzdnych ekranu
  piszilcd(0x0e);  //wcz LCD, wcz kursor bez mrugania
  for(i=0;i<8;i++)
  {
   lcdy(i);           //umie kursor
   czekaj(500*tau);
   czysclcd();
   lcdy(15-i);        //umie kursor
   czekaj(500*tau);
   czysclcd();
  }

//>>>>>>>>>>>>   EFEKT 3   <<<<<<<<<<<<<<<
//mruganie tekstem wywietlanym na LCD
  czysclcd();
  pisztekst(tekst2); //wywietl tekst na LCD

  for(i=0;i<20;i++)
  {
   piszilcd(0x08);   //wycz LCD
   czekaj(300*tau);
   piszilcd(0x0c);   //wcz LCD
   czekaj(300*tau);
   if(!(PIND&0x02))
   {
    break;           //zakoczenie ptli po naciniciu SW4
   }
  }
  czysclcd();
  czekaj(200*tau);
 }
}

