/****************************************************************/
/* wiczenie 5 - Obsuga klawiatury matrycowej z wykorzystaniem */
/*               przerwa timera0                               */
/*             - wykorzystanie wywietlacza LCD 16*2            */
/*             - prosta gra zrcznociowa                       */
/* J.D. '2003                                                   */
/****************************************************************/
#include <io.h>
#include <interrupt.h>
#include <signal.h>

#define lcd_rs 2       //definicja bitu portu dla linii RS
#define lcd_e 3        //definicja bitu portu dla linii E
#define tau0 247       //staa czasowa timera0
#define vliczt0 113    //staa wpisywana do licznika wej do
                       //do przerwania timera0
#define vlkursor 10    //warto wpisywana do zmiennej lkursor

//********************  zmienne globalne  *******************
unsigned char liczt0;  //Licznik wej do przerwania timera0.
                       //Klawisz jest badany, gdy liczt0=0
unsigned char lkursor; //licznik wej do przerwania timera0 dla 
                       //ustawienia flagi zmiany pooenia "lisa"       
volatile unsigned char kodklaw;  //kod nacinitego klawisza;
volatile unsigned char fklaw;    //flaga wykrycia nacinicia klawisza;
volatile unsigned char fkursor;  //flaga zmiany pooenia "lisa"
unsigned char wiersz=0;          //pozycja umieszczenia znaku na LCD
unsigned char kolumna=0;         //pozycja umieszczenia znaku na LCD

void czekaj(unsigned long zt) //procedura wytracania czasu
{
#define tau 10.38      //przybliony przelicznik argumentu na ms
 unsigned char zt1;
 
 for(;zt>0;zt--)
 {
  for(zt1=255;zt1!=0;zt1--);
 }
}

SIGNAL (SIG_OVERFLOW0)  //obsuga przerwania od przepenienia timera0
{
 unsigned char kkolumna,kwiersz; //zmienne pomocnicze
 TCNT0=tau0;            //odwie sta czasow w TCNT0
 
 if(--liczt0==0)        //czy ju czyta klawisze?
 {                      //tak
  for(kkolumna=0xfe;kkolumna!=0xfb;kkolumna=(kkolumna<<1)+1)
  {
   PORTD=(PORTD|0x03)&(kkolumna);    //podaj "0" na lini sterujc aktywn kolumn 
   kwiersz=PIND&0x0f;  //pozostaw tylko linie obsugujce klawiatur
   if((kwiersz&0x0c)^0x0c)    //czy jest odpowied na jakiej linii wierszy?
   {                          //tak, ustaw flag gotowoci klawiatury
    fklaw=1;
    break;                    //i zakocz przepatrywanie
   }
  }
  if(fklaw)
  {
   kodklaw=kwiersz;    //jeli wykryto wcinicie klawisza, podaj jego kod       
  }
  else
  {
   kodklaw=0xff;       //jeli nie, ustaw kod neutralny
  }
  liczt0=vliczt0;      //odwie stan liczt0 
  if(--lkursor==0)     //czy mona ustawi flag zmiany "lisa"?
  {
   fkursor=1;          //tak (mino ok. 300ms)
   lkursor=vlkursor;   //odwie stan lkursor
  }
 }
}

void piszilcd(unsigned char instr) //zapisz instrukcj sterujc do LCD
{
 cbi(PORTB,lcd_rs);
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|(instr&0xf0);  //przygotuj starszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD ok. 100us
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|((instr&0x0f)<<4);  //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD ok. 100us
}

void piszdlcd(char dana)           //zapisz dan do LCD
{
 sbi(PORTB,lcd_rs);
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|(dana&0xf0);   //przygotuj starszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|((dana&0x0f)<<4); //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);                 //impuls strobujcy
 czekaj(10L);                      //czekaj na gotowo LCD
}

void czysclcd(void)                //czy ekran
{
 piszilcd(0x01);      //polecenie czyszczenia ekranu dla kontrolera LCD
 czekaj(1.64*tau);    //rozkaz 0x01 wykonuje si 1.64ms
 wiersz=0;
 kolumna=0;
}

void lcdxy(unsigned char w, unsigned char k)    //ustaw wsprzdne kursora
{
 piszilcd((w*0x40+k)|0x80);  //standardowy rozkaz sterownika LCD
                             //ustawiajcy kursor w okrelonych wsprzdnych
}

void piszznak(char znak)    //procedura umieszcza znak na wywietlaczu
{
 piszdlcd(znak);            //wywietl znak na LCD
}

unsigned char incwiersz(unsigned char w)  //inkrementuj wiersz LCD
{
 return w==1?0:1;
}

unsigned char decwiersz(unsigned char w)  //dekrementuj wiersz LCD
{
 return w==0?1:0;
}

unsigned char inckolumna(unsigned char k) //inkrementuj kolumn LCD 
{
 return k==15?0:++k;
}

unsigned char deckolumna(unsigned char k) //dekrementuj kolumn LCD
{
 return k==0?15:--k;
}

int main(void)         //program gwny
{
 unsigned char r,i;    //zmienne pomocnicze
 unsigned char xg=0;   //kolumna wywietlania kursora gracza
 unsigned char yg=0;   //wiersz wywietlania kursora gracza
 
 liczt0=vliczt0;
 lkursor=vlkursor;
 PORTD=0xff;           //z podciganiem
 PORTB=0x03;           //z podciganiem wej
 DDRD=0x03;            //PD0-PD1 wy, pozostae wejcia
 DDRB=0xff;            //PORTB - wy
 TIMSK=1<<TOIE0;       //zezwolenie na przerwania od TC0
 TCNT0=tau0;           //wpisz sta czasowa do TCNT0
 TCCR0=4;              //preskaler XTAL/256,
                       //kwant mierzonego czasu = 32<m>s
 fklaw=0;
  //>>>>>>>>>>>>>>>>>>   inicjacja wywietlacza LCD   <<<<<<<<<<<<<<<<<<<<<<
 czekaj(45*tau);        //opnienie ok. 45ms dla ustabilizowania si napicia
                        //zasilania LCD (katalogowo min. 15 ms)
                        //lcd_rs ju wczeniej byo ustawione w stanie "0"
 for(i=0;i<3;i++)       //3-krotne wysanie 3-
 {
  sbi(PORTB,lcd_e);
  PORTB=(PORTB&0x0f)|0x30; //wylij 3- do LCD
  asm("nop");
  asm("nop");
  asm("nop");
  cbi(PORTB,lcd_e);
  czekaj(5*tau);      //ok. 5ms
 }
 sbi(PORTB,lcd_e);
 PORTB=(PORTB&0x0f)|0x20;  //wylij 2- do LCD
 asm("nop");               //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 cbi(PORTB,lcd_e);     //impuls strobujcy
 czekaj(10L); 
 piszilcd(0x28);       //interfejs 4-bitowy, 2 linie, znak 5x7
 piszilcd(0x08);       //wycz LCD, wycz kursor, wycz mruganie
 piszilcd(0x01);       //czy LCD
 czekaj(1.64*tau);     //wymagane dla instrukcji czyszczenia ekranu opnienie
 piszilcd(0x06);       //bez przesuwania w prawo
 piszilcd(0x0c);       //wcz LCD, bez kursora, bez mrugania
 sei();                //odblokuj globalne przerwania
 i=0;

 piszznak('S');        //wywietl napis "Start"
 piszznak('t');
 piszznak('a');
 piszznak('r');
 piszznak('t');
  
 do 
 {
  i++;                 //zmienna "i" bdzie wykorzystana pniej do zainicjowania 
                       //generatora pseudolosowego
 }while(!fklaw);       //nacinicie klawisza zapewnia losowy start generatora
 srand(i);             //inicjuj generator liczb pseudolosowych
 czysclcd();           //czy ekran
 piszznak('+');        //umie znak gracza na ekranie
 
 while(1)              //gwna ptla programu
 {
  if(fkursor)          //czy mona przemieci "lisa"?
  {
   lcdxy(wiersz,kolumna);  //tak...
   if((wiersz==yg)&&(kolumna==xg))
   {
    piszznak('+');      //wyma star pozycj pozostawiajc gracza
   }
   else
   {
    piszznak(' ');      //wyma star pozycj
   }   
   r=rand();           //losuj nowe pooenie
   if((r<37)&&(r>18))  //wylicz now pozycj
   {
    wiersz=incwiersz(wiersz);
   }
   if(r<19)
   {
    wiersz=decwiersz(wiersz);
   }
   if(((r>12)&&(r<19))||((r>30)&&(r<147)))
   {
    kolumna=inckolumna(kolumna);
   }
   if((r<7)||((r<25)&&(r>18))||(r>146))
   {
    kolumna=deckolumna(kolumna);
   }   
   lcdxy(wiersz,kolumna);     //ustaw nowe wsprzdne
   piszznak('.');      //umie "lisa" w nowym pooeniu
   fkursor=0;
  }
  
  if(fklaw)
  {                    //wykryto nacinicie klawisza
   fklaw=0;
   lcdxy(yg,xg);       //ustaw wsprzdne znaku gracza...
   piszznak(' ');      //... i wyma go 
   switch (kodklaw)    //reakcja na klawisz
   {
    case 0x09:  xg=inckolumna(xg);      //SW4 - w prawo z zawijaniem
                break;
    case 0x05:  xg=deckolumna(xg);      //SW3 - w lewo z zawijaniem
                break;
    case 0x06:  yg=1;                   //SW2 - na d bez zawijania 
                break;            
    case 0x0a:  yg=0;                   //SW1 - do gry bez zawijania
   }                                
   lcdxy(yg,xg);       //ustaw nowe wsprzdne znaku gracza
   if((wiersz==yg)&&(kolumna==xg))
   {
    piszznak('*');     //znak, jeli trafiony
    czekaj(200);        //przytrzymaj chwil na ekranie
   }
   else
   {
    piszznak('+');     //znak jeli "pudo"
   }
  }
 } 
}

